""" Generator functions related to trial and test functions and the accumulation loop"""

from dune.perftool.generation import domain, iname, pymbolic_expr, symbol, globalarg, function_mangler, constantarg, get_global_context_value
from dune.perftool.ufl.modified_terminals import ModifiedArgumentDescriptor
from dune.perftool.pdelab import (name_index,
                                  restricted_name,
                                  )
from dune.perftool.pdelab.basis import (evaluate_trialfunction,
                                        evaluate_trialfunction_gradient,
                                        lfs_iname,
                                        name_basis,
                                        name_basis_gradient,
                                        name_lfs_bound,
                                        )
from dune.perftool import Restriction
from pymbolic.primitives import Call, Subscript, Variable


@symbol
def name_testfunction_gradient(element, restriction):
    return name_basis_gradient(element, restriction)


@symbol
def name_testfunction(element, restriction):
    it = get_global_context_value("integral_type")
    if it == 'exterior_facet':
        restriction = Restriction.NEGATIVE
    return name_basis(element, restriction)


@symbol
def name_trialfunction_gradient(element, restriction, component):
    rawname = "gradu" + "_".join(str(c) for c in component)
    name = restricted_name(rawname, restriction)
    evaluate_trialfunction_gradient(element, name, restriction, component)
    return name


@symbol
def name_trialfunction(element, restriction, component):
    rawname = "u" + "_".join(str(c) for c in component)
    name = restricted_name(rawname, restriction)
    evaluate_trialfunction(element, name, restriction, component)
    return name


@symbol
def name_testfunctionspace(restriction):
    return restricted_name("lfsv", restriction)


@symbol
def name_trialfunctionspace(restriction):
    return restricted_name("lfsu", restriction)


@symbol
def type_testfunctionspace():
    return "LFSV"


@symbol
def type_trialfunctionspace():
    return "LFSU"


@symbol
def name_coefficientcontainer(restriction):
    name = restricted_name("x", restriction)
    from dune.perftool.pdelab.basis import name_lfs_bound, lfs_iname
    return name


@function_mangler
def create_function_mangler(container):
    def _mangler(kernel, name, dtypes):
        if name == container:
            import loopy
            return loopy.types.NumpyType("int"), container

    return _mangler


@pymbolic_expr
def pymbolic_coefficient(lfs, index, restriction):
    container = name_coefficientcontainer(restriction)
    create_function_mangler(container)
    import loopy
    constantarg(lfs, dtype=loopy.types.NumpyType("str"))
    from dune.perftool.pymbolic import VerbatimVariable
    return Call(VerbatimVariable(container), (VerbatimVariable(lfs), Variable(index),))


@symbol
def type_coefficientcontainer():
    return "X"


def name_argumentspace(ma):
    from ufl.classes import Argument, Coefficient
    if isinstance(ma.argexpr, Argument):
        if ma.argexpr.number() == 0:
            return name_testfunctionspace(ma.restriction)
        if ma.argexpr.number() == 1:
            return name_trialfunctionspace(ma.restriction)
    if isinstance(ma.argexpr, Coefficient):
        assert ma.argexpr.count() == 0
        return name_trialfunctionspace(ma.restriction)
    # We should never encounter an argument other than 0 or 1
    assert False


def name_argument(ma):
    if ma.argexpr.number() == 0:
        return name_testfunction(ma)
    if ma.argexpr.number() == 1:
        return name_trialfunction(ma)
    # We should never encounter an argument other than 0 or 1
    assert False


@symbol
def name_jacobian(restriction1, restriction2):
    # Restrictions may only differ if NONE
    if (restriction1 == Restriction.NONE) or (restriction2 == Restriction.NONE):
        assert restriction1 == restriction2

    return restricted_name(restricted_name("jac", restriction1), restriction2)


@symbol
def type_jacobian():
    return "J"


@symbol
def name_residual(restriction):
    return restricted_name("r", restriction)


@symbol
def type_residual():
    return "R"


def name_accumulation_variable(restrictions=(Restriction.NONE,)):
    from dune.perftool.generation import get_global_context_value
    ft = get_global_context_value("form_type")
    if ft == 'residual':
        return name_residual(*restrictions)
    if ft == 'jacobian':
        return name_jacobian(*restrictions)
    assert False


def type_accumulation_variable():
    from dune.perftool.generation import get_global_context_value
    ft = get_global_context_value("form_type")
    if ft == 'residual':
        return type_residual()
    if ft == 'jacobian':
        return type_jacobian()
    assert False
